<?php

declare(strict_types=1);

namespace Drupal\lms;

use Drupal\Core\Cache\Cache;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Security\Attribute\TrustedCallback;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Drupal\lms\Entity\Bundle\Course;
use Drupal\lms\Entity\LessonInterface;
use Symfony\Component\DependencyInjection\Attribute\Autowire;

/**
 * Builds the course navigation block.
 */
class BlockBuilder {

  private const CACHE_PREFIX = 'steps_block_data_cache_';

  /**
   * Static cache of user lesson statuses.
   */
  private ?array $lessonStatuses = NULL;

  /**
   * Static cache of route parameters.
   */
  private ?array $routeParameters = NULL;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    protected readonly AccountInterface $currentUser,
    #[Autowire(service: 'cache.default')]
    protected readonly CacheBackendInterface $cache,
    protected readonly TrainingManager $trainingManager,
    protected readonly RouteMatchInterface $currentRouteMatch,
    protected readonly EntityTypeManagerInterface $entityTypeManager,
  ) {
  }

  /**
   * Get route parameter.
   */
  private function getRouteParameter(string $parameter): ?int {
    if ($this->routeParameters === NULL) {
      $this->routeParameters = [];
      foreach (['lesson_delta', 'activity_delta'] as $param) {
        $this->routeParameters[$param] = $this->currentRouteMatch->getParameter($param);
      }
    }
    return \array_key_exists($parameter, $this->routeParameters) ? (int) $this->routeParameters[$parameter] : NULL;
  }

  /**
   * Set lesson statuses.
   */
  private function setLessonStatuses(Course $group, array $lesson_ids): void {
    if ($this->lessonStatuses !== NULL) {
      return;
    }
    $this->lessonStatuses = [];

    $course_status = $this->trainingManager->loadCourseStatus($group, $this->currentUser, [
      'current' => TRUE,
    ]);
    if ($course_status === NULL) {
      return;
    }

    $lesson_statuses = $this->trainingManager->loadLessonStatusMultiple($course_status->id(), $lesson_ids);
    foreach ($lesson_statuses as $lesson_status) {
      $this->lessonStatuses[$lesson_status->get('lesson')->target_id] = $lesson_status;
    }
  }

  /**
   * Get data structure for this training.
   */
  private function getStructure(Course $group): array {
    $cid = self::CACHE_PREFIX . $group->id();
    $cache = $this->cache->get($cid);
    if ($cache !== FALSE) {
      $step_data = $cache->data;
    }
    else {
      $step_data = $this->trainingManager->getOrderedLessons($group);
      $invalidate_tags = ['group:' . $group->id()];
      foreach (\array_keys($step_data) as $lesson_id) {
        $invalidate_tags[] = 'lms_lesson:' . $lesson_id;
      }
      $this->cache->set($cid, $step_data, Cache::PERMANENT, $invalidate_tags);
    }

    return $step_data;
  }

  /**
   * Builds the lesson navigation render array.
   */
  public function build(Course $group, AccountInterface $user): array {
    $build = [
      'course' => [
        '#cache' => [
          'max-age' => 0,
        ],
        'heading' => [
          '#type' => 'html_tag',
          '#tag' => 'h2',
          '#value' => $group->label(),
          '#attributes' => [
            'class' => [
              'lms-course-navigation',
            ],
          ],
        ],
        'lessons' => [
          '#type' => 'html_tag',
          '#tag' => 'ol',
          '#attributes' => [
            'class' => [
              'lms-course-navigation',
            ],
          ],
        ],
      ],
    ];
    $structure = $this->getStructure($group);
    foreach ($structure as $lesson_id => $step_data) {
      $is_current = ($this->getRouteParameter('lesson_delta') === $step_data['delta']);
      $lesson = [
        '#lazy_builder' => ['lms.block_builder:buildLesson', [$lesson_id, $is_current, $group->id(), $user->id()]],
        '#create_placeholder' => TRUE,
      ];
      if (!$is_current) {
        $lesson['#cache'] = [
          '#keys' => ['lms-lesson', $lesson_id, $group->id()],
          '#contexts' => ['user'],
        ];
      }
      else {
        $lesson['#cache']['max-age'] = 0;
      }
      $build['course']['lessons'][$lesson_id] = $lesson;
    }
    return $build;
  }

  /**
   * Builds the render array for a lesson.
   */
  #[TrustedCallback]
  public function buildLesson(int $lesson_id, bool $is_current, int $group_id, int $user_id): array {
    $lesson_storage = $this->entityTypeManager->getStorage('lms_lesson');
    $lesson = $lesson_storage->load($lesson_id);
    $course_storage = $this->entityTypeManager->getStorage('group');
    $group = $course_storage->load($group_id);
    assert($group instanceof Course);

    $structure = $this->getStructure($group);
    $step_data = $structure[$lesson_id];
    $lesson_id = (string) $lesson_id;
    $this->setLessonStatuses($group, \array_keys($structure));
    $is_current = ($this->getRouteParameter('lesson_delta') === $step_data['delta']);
    $build = [
      '#type' => 'html_tag',
      '#tag' => 'li',
      '#attributes' => [
        'class' => [
          'lesson-activities',
          $is_current ? 'current' : '',
        ],
      ],
      '#started' => FALSE,
      '#finished' => FALSE,
      '#evaluated' => FALSE,
      '#passed' => FALSE,
      '#current' => $is_current,
    ];
    if (\array_key_exists($lesson_id, $this->lessonStatuses)) {
      $build['#started'] = TRUE;
      $lesson_status = $this->lessonStatuses[$lesson_id];
      $lesson = $lesson_status->get('lesson')->entity;

      $cacheable_metadata = new CacheableMetadata();
      $cacheable_metadata->addCacheableDependency($lesson_status);
      $cacheable_metadata->addCacheableDependency($lesson);
      $cacheable_metadata->applyTo($build);
      if ($lesson_status->isFinished()) {
        $build['#finished'] = TRUE;
        if ($lesson_status->getScore() >= $lesson_status->getRequiredScore()) {
          $build['#passed'] = TRUE;
        }
      }
      $build['#evaluated'] = $lesson_status->isEvaluated();
    }
    $classes = [
      'lesson-activities',
      $build['#started'] ? 'started' : 'not-started',
      $build['#finished'] ? 'finished' : 'not-finished',
      $build['#evaluated'] ? 'evaluated' : 'not-evaluated',
      $build['#passed'] ? 'passed' : 'not-passed',
    ];

    $build['#attributes']['class'] = $classes;

    $lesson_url = Url::fromRoute('lms.group.answer_form', [
      'group' => $group->id(),
      'lesson_delta' => $step_data['delta'],
    ]);

    $build['title'] = [
      '#type' => 'link',
      '#title' => $step_data['label'],
      '#url' => $lesson_url,
      '#options' => [
        'attributes' => [
          'class' => [
            'lesson-title',
            $is_current ? 'current' : '',
          ],
        ],
      ],
    ];

    $build['activities'] = [
      '#type' => 'html_tag',
      '#tag' => 'ol',
      '#current' => $is_current,
      '#cache' => [
        'max-age' => $is_current ? 0 : Cache::PERMANENT,
      ],
    ];

    $cacheable_metadata = new CacheableMetadata();
    $cacheable_metadata->addCacheableDependency($lesson);
    $cacheable_metadata->applyTo($build);

    if (\array_key_exists($lesson_id, $this->lessonStatuses)) {
      $this->applyLessonStatusData($build, $step_data, $group->id(), $lesson_id, $step_data['delta']);
    }

    return ['lesson' => $build];
  }

  /**
   * Merge activity renderable to activities renderable.
   */
  private function mergeActivityRenderables(array &$activities, array $activities_data, string $group_id, int $step_delta): void {
    foreach ($activities_data as $activity_id => $activity_data) {
      $is_current = $activity_data['current'] ?? FALSE;
      $activity_url = Url::fromRoute('lms.group.answer_form', [
        'group' => $group_id,
        'lesson_delta' => $step_delta,
        'activity_delta' => $activity_data['delta'],
      ]);
      $classes = [
        $activity_data['answered'] ? 'answered' : 'not-answered',
        $activity_data['max_score_achieved'] ? 'full-marks' : 'not-full-marks',
        $activity_data['evaluated'] ? 'evaluated' : 'not-evaluated',
        $is_current ? 'current' : '',
      ];

      $activities[$activity_id] = [
        '#type' => 'html_tag',
        '#tag' => 'li',
        '#attributes' => [
          'class' => $classes,
        ],
        'link' => [
          '#type' => 'link',
          '#title' => $activity_data['label'],
          '#url' => $activity_url,
          '#options' => [
            'attributes' => [
              'class' => array_merge($classes, [
                'activity-title',
              ]),
            ],
          ],
        ],
        '#answered' => $activity_data['answered'],
        '#max_score_achieved' => $activity_data['max_score_achieved'],
        '#max_score' => $activity_data['max_score'],
        '#score' => $activity_data['score'],
        '#evaluated' => $activity_data['evaluated'],
        '#current' => $is_current,
        '#activity' => $activity_data['activity'],
      ];
      $cacheable_metadata = new CacheableMetadata();
      $cacheable_metadata->addCacheableDependency($activity_data['activity']);
      $cacheable_metadata->applyTo($activities[$activity_id]);
    }
  }

  /**
   * Apply lesson status data to lesson renderable.
   */
  private function applyLessonStatusData(array &$build, array $step_data, string $group_id, string $lesson_id, int $step_delta): void {
    if (!\array_key_exists($lesson_id, $this->lessonStatuses)) {
      return;
    }
    $lesson_status = $this->lessonStatuses[$lesson_id];
    $lesson = $lesson_status->getLesson();
    $cacheable_metadata = new CacheableMetadata();
    $cacheable_metadata->addCacheableDependency($lesson_status);
    $cacheable_metadata->applyTo($build);
    $activities_data = [];
    $activity_ids = [];
    $activities_field = $lesson_status->get(LessonInterface::ACTIVITIES);
    foreach ($activities_field as $activity_item) {
      $activity_ids[$activity_item->target_id] = $activity_item->target_id;
    }
    $answers = $this->trainingManager->getAnswersByActivityId($lesson_status, $activity_ids);

    foreach ($activities_field as $activity_delta => $activity_item) {
      $activity = $activity_item->entity;
      if ($activity === NULL) {
        continue;
      }
      $answered = FALSE;
      $evaluated = FALSE;
      $maximum_score_achieved = FALSE;
      $score = NULL;
      if (array_key_exists($activity->id(), $answers)) {
        $answered = TRUE;
        if ($answers[$activity->id()]->isEvaluated()) {
          $evaluated = TRUE;
        }
        $score = $answers[$activity->id()]->getScore();
      }
      $maximum_score = $this->trainingManager->getActivityMaxScore($lesson, $activity);
      if ($score >= $maximum_score) {
        $maximum_score_achieved = TRUE;
      }
      $activities_data[$activity->id()] = [
        'answered' => $answered,
        'evaluated' => $evaluated,
        'max_score_achieved' => $maximum_score_achieved,
        'max_score' => $this->trainingManager->getActivityMaxScore($lesson, $activity),
        'score' => $score,
        'label' => $activity->label(),
        'current' => $build['#current'] && ($activity_delta === $this->getRouteParameter('activity_delta')),
        'activity' => $activity,
        'delta' => $activity_delta,
      ];
    }
    $this->mergeActivityRenderables($build['activities'], $activities_data, $group_id, $step_delta);
  }

}
